"""Colorizing utils"""

import re
import typing as t

from ..config.constants import C

__all__ = [
    "Color",
]


class Color:
    """Text color wrapping"""

    _COLOR_RE: t.Pattern = re.compile(r"([^\n]+)")

    @classmethod
    def gray(cls, message: str) -> str:
        """Make a string gray"""
        return cls._add_formatting(message, 90)

    @classmethod
    def red(cls, message: str) -> str:
        """Make a string red"""
        return cls._add_formatting(message, 31)

    @classmethod
    def green(cls, message: str) -> str:
        """Make a string green"""
        return cls._add_formatting(message, 32)

    @classmethod
    def yellow(cls, message: str) -> str:
        """Make a string yellow"""
        return cls._add_formatting(message, 33)

    @classmethod
    def blue(cls, message: str) -> str:
        """Make a string blue"""
        return cls._add_formatting(message, 34)

    @classmethod
    def bold(cls, message: str) -> str:
        """Make a string bold"""
        return cls._add_formatting(message, 1)

    @classmethod
    def _add_formatting(cls, message: str, code: int) -> str:
        if not C.USE_COLOR:
            return message
        return cls._COLOR_RE.sub(f"\u001b[{code}m\\1\u001b[0m", message)



"""Loading dataclasses from dicts"""

import enum
import logging
import pathlib
import typing as t

from .exceptions import RootTypeUnionMatchError
from .loader import DataClassLoader

T = t.TypeVar("T")

__all__ = [
    "from_dict",
    "get_data_class_by_data_signature",
]

logger = logging.getLogger(__name__)


def from_dict(data_type: type[T], data: dict[str, t.Any]) -> T:
    """Create a dataclass instance from dict"""
    class_loader = DataClassLoader(
        validate_types=True,
        cast_types_list=[
            enum.Enum,
            pathlib.Path,
        ],
    )
    return class_loader.from_dict(data_type=data_type, data=data)


def get_data_class_by_data_signature(data_type: type[T], data: dict[str, t.Any]) -> type[T]:
    """Select a dataclass from possible union type"""
    class_loader = DataClassLoader(validate_types=False)
    if not class_loader._is_union(data_type):  # pylint: disable=protected-access
        class_loader.from_dict(data_type=data_type, data=data)
        return data_type

    union_matches: list[type[T]] = []
    inner_types: tuple[type[T], ...] = class_loader._extract_generic_args(data_type)  # pylint: disable=protected-access
    for inner_type in inner_types:
        try:
            source_class = get_data_class_by_data_signature(data_type=inner_type, data=data)
        except Exception as e:
            logger.debug(f"Type {inner_type}: {e}")
        else:
            union_matches.append(source_class)
    if len(union_matches) != 1:
        raise RootTypeUnionMatchError(types=inner_types, keys=sorted(data)) from None
    return union_matches[0]



"""All intercepted errors"""

__all__ = [
    "ExecutionFailed",
    "ActionRenderError",
    "PendingActionUnresolvedOutcomeError",
    "ActionRenderRecursionError",
    "RestrictedBuiltinError",
    "ActionRunError",
    "BaseError",
    "LoadError",
    "IntegrityError",
    "SourceError",
    "InteractionError",
    "ActionArgumentsLoadError",
    "YAMLStructureError",
    "AutoStrategyCycleError",
]

import pathlib


class ExecutionFailed(Exception):
    """Some steps failed"""


class ActionRenderError(Exception):
    """Action rendering failed"""


class PendingActionUnresolvedOutcomeError(ActionRenderError):
    """Action rendering failed due to unresolved outcome"""

    def __init__(self, action_name: str):
        self.action_name: str = action_name
        super().__init__(f"Action {action_name!r} has not finished yet, therefore its outcomes are unresolved")


class AutoStrategyCycleError(Exception):
    """Automatic strategy discovered cyclic dependencies"""


class ActionArgumentsLoadError(Exception):
    """Action arguments loading failed"""


class ActionRenderRecursionError(ActionRenderError):
    """Action recursion depth exceeded"""


class RestrictedBuiltinError(Exception):
    """Action rendering access to a restricted builtin function"""


class ActionRunError(Exception):
    """Action execution failed"""


class BaseError(Exception):
    """Common base to catch in CLI"""

    CODE: int = 101


class LoadError(BaseError):
    """Loader regular exception during load process"""

    CODE: int = 102

    def __init__(self, message: str, stack: tuple[pathlib.Path, ...]) -> None:
        self.message: str = message
        self.stack: tuple[pathlib.Path, ...] = stack
        text: str = message
        if stack:
            text += f"\n  Sources stack: {' -> '.join(str(path) for path in stack)}"
        super().__init__(text)


class IntegrityError(BaseError):
    """Workflow structure error"""

    CODE: int = 103


class SourceError(BaseError):
    """Source file not recognized"""

    CODE: int = 104


class InteractionError(BaseError):
    """Can't interact with a display"""

    CODE: int = 105


class YAMLStructureError(BaseError):
    """Custom tags structure error"""

    CODE: int = 106



"""Lazy-loaded constants implementations"""

import os
import sys
import typing as t
from io import UnsupportedOperation
from pathlib import Path

from . import base
from .cli import get_cli_option
from .helpers import class_from_module
from .runner import TEMP_DIR_CONTEXT
from ...strategy.base import BaseStrategy
from ...types import (
    LoaderClassType,
    StrategyClassType,
    DisplayClassType,
)

__all__ = [
    "LogLevel",
    "LogFile",
    "RcFile",
    "ContextDirectory",
    "InteractiveMode",
    "DependencyDefaultStrictness",
    "WorkflowSourceFile",
    "WorkflowLoaderClass",
    "InternalDisplayClass",
    "ExternalDisplayClass",
    "StrategyClass",
    "UseColor",
    "DefaultShellExecutable",
    "ShellInjectYieldFunction",
    "StrictOutcomesRendering",
    "ActionClassDirectories",
    "ExternalPythonModulesPaths",
    "TemplarModulesWhitelist",
]


class LogLevel(base.ConstantBase[str]):
    """Specifies the level for the logging subsystem."""

    ENVIRONMENT_VARIABLE_NAME = "GRANA_LOG_LEVEL"
    COMMAND_LINE_OPTION_NAME = "log_level"
    RC_PARAMETER_NAME = "log_level"
    DEFAULT = "ERROR"

    def cast(self, value: t.Any) -> str:
        log_levels_normalization_map: dict[str, str] = {
            "0": "ERROR",
            "1": "WARNING",
            "2": "INFO",
            "3": "DEBUG",
        }
        allowed_log_levels: list[str] = list(log_levels_normalization_map.values()) + list(log_levels_normalization_map)
        if value not in allowed_log_levels:
            raise ValueError(f"{value!r} is not a valid log level (expected one of: {allowed_log_levels})")
        return log_levels_normalization_map.get(value, value)


class LogFile(base.ConstantPath[t.Optional[Path]]):
    """Specifies the log file path."""

    ENVIRONMENT_VARIABLE_NAME = "GRANA_LOG_FILE"
    COMMAND_LINE_OPTION_NAME = "log_file"
    RC_PARAMETER_NAME = "log_file"
    DEFAULT = None


class RcFile(base.ConstantPath[Path]):
    """Specifies the runtime configuration file location."""

    ENVIRONMENT_VARIABLE_NAME = "GRANA_RC_FILE"

    def default(self) -> Path:
        """`.granarc` file in the current working directory"""
        return Path().resolve() / ".granarc"


class RunnerTemporaryDirectory(base.ConstantPath[t.Optional[Path]]):
    """Holds a runner-scoped temporary directory path"""

    def default(self) -> t.Optional[Path]:
        """Obtain from runner context"""
        return TEMP_DIR_CONTEXT.get()


class ContextDirectory(base.ConstantPath[Path]):
    """A directory that is used to resolve all relative paths."""

    def default(self) -> Path:
        """Current working directory"""
        return Path().resolve()


class InteractiveMode(base.ConstantBool):
    """Specifies whither to run plan interaction phase or not."""

    COMMAND_LINE_OPTION_NAME = "interactive"
    DEFAULT = False


class DependencyDefaultStrictness(base.ConstantBool):
    """Default strictness for action dependencies."""

    ENVIRONMENT_VARIABLE_NAME = "GRANA_STRICT_DEPENDENCIES"
    RC_PARAMETER_NAME = "strict"
    WORKFLOW_CONFIG_PARAMETER_NAME = "strict"
    DEFAULT = False


class WorkflowSourceFile(base.ConstantPath[t.Optional[Path]]):
    """Workflow source file path.
    When not set, `grana.yml`/`grana.yaml` are being looked for in the working directory.
    If set to `-`, then standard input stream is used as the source."""

    ENVIRONMENT_VARIABLE_NAME = "GRANA_WORKFLOW_FILE"
    RC_PARAMETER_NAME = "workflow_file"
    DEFAULT = None

    def from_cli_option(self) -> t.Optional[Path]:
        """Positional argument to `grana run` and `grana validate`."""
        cli_arg_name: str = "workflow_file"
        if (cli_arg_value := get_cli_option(cli_arg_name)) is None:
            raise base.Inapplicable
        self.logger.debug(f"Defined CLI positional argument {cli_arg_name!r} is accessed by {self._name!r}")
        return self.cast(cli_arg_value)


class WorkflowLoaderClass(base.ConstantBase[t.Optional[LoaderClassType]]):
    """May point to a file containing a WorkflowLoader class definition,
    which will replace the default implementation."""

    ENVIRONMENT_VARIABLE_NAME = "GRANA_WORKFLOW_LOADER_SOURCE_FILE"
    RC_PARAMETER_NAME = "workflow_loader_source_file"
    DEFAULT = None

    def cast(self, value: str) -> LoaderClassType:
        return t.cast(
            LoaderClassType,
            class_from_module(
                source_path=Path(value),
                class_name="WorkflowLoader",
                submodule_name="workflow.loader",
            ),
        )


class InternalDisplayClass(base.ConstantBase[DisplayClassType]):
    """Selects a display by name from the bundled list."""

    ENVIRONMENT_VARIABLE_NAME = "GRANA_DISPLAY_NAME"
    RC_PARAMETER_NAME = "display_name"
    COMMAND_LINE_OPTION_NAME = "display"

    def cast(self, value: str) -> DisplayClassType:
        from ...display.default import KNOWN_DISPLAYS

        try:
            return KNOWN_DISPLAYS[value]
        except Exception:
            raise ValueError(f"Display name should be one of: {sorted(KNOWN_DISPLAYS)}. Got {value!r}") from None

    def default(self) -> DisplayClassType:
        """`prefix`"""
        from ...display.default import PrefixDisplay

        return PrefixDisplay


class ExternalDisplayClass(base.ConstantBase[t.Optional[DisplayClassType]]):
    """May point to a file containing a Display class definition, which will replace the default implementation."""

    ENVIRONMENT_VARIABLE_NAME = "GRANA_DISPLAY_SOURCE_FILE"
    RC_PARAMETER_NAME = "display_source_file"
    DEFAULT = None

    def cast(self, value: str) -> DisplayClassType:
        return t.cast(
            DisplayClassType,
            class_from_module(
                source_path=Path(value),
                class_name="Display",
                submodule_name="display",
            ),
        )


class StrategyClass(base.ConstantBase[StrategyClassType]):
    """Specifies the execution strategy."""

    ENVIRONMENT_VARIABLE_NAME = "GRANA_STRATEGY_NAME"
    COMMAND_LINE_OPTION_NAME = "strategy"
    WORKFLOW_CONFIG_PARAMETER_NAME = "strategy"
    RC_PARAMETER_NAME = "strategy"

    def cast(self, value: str) -> StrategyClassType:
        return BaseStrategy.get_strategy_class_by_name(value)

    def default(self) -> StrategyClassType:
        """`auto`"""
        from ...strategy.impl import AutoStrategy

        return AutoStrategy


class UseColor(base.ConstantBool):
    """When specified, this will force the colored or non-colored output, according to the setting."""

    ENVIRONMENT_VARIABLE_NAME = "GRANA_FORCE_COLOR"
    RC_PARAMETER_NAME = "force_color"

    def default(self) -> bool:
        """Depending on if a TTY is allocated."""
        try:
            return os.isatty(sys.stdout.fileno())
        except UnsupportedOperation:
            return False


class DefaultShellExecutable(base.ConstantBase[str]):
    """Specifies which shell executable should be used by the shell action by default."""

    ENVIRONMENT_VARIABLE_NAME = "GRANA_DEFAULT_SHELL_EXECUTABLE"
    RC_PARAMETER_NAME = "default_shell_executable"
    WORKFLOW_CONFIG_PARAMETER_NAME = "default_shell_executable"
    DEFAULT = "/bin/sh"


class ShellInjectYieldFunction(base.ConstantBool):
    """When set to True, all shell-related actions will inject the yield_outcome function definition."""

    ENVIRONMENT_VARIABLE_NAME = "GRANA_SHELL_INJECT_YIELD_FUNCTION"
    RC_PARAMETER_NAME = "shell_inject_yield_function"
    WORKFLOW_CONFIG_PARAMETER_NAME = "shell_inject_yield_function"
    DEFAULT = True


class StrictOutcomesRendering(base.ConstantBool):
    """When set to True, rendering a missing outcome key will result in an error instead of an empty string."""

    ENVIRONMENT_VARIABLE_NAME = "STRICT_OUTCOMES_RENDERING"
    RC_PARAMETER_NAME = "strict_outcomes_rendering"
    WORKFLOW_CONFIG_PARAMETER_NAME = "strict_outcomes_rendering"
    DEFAULT = True


class ActionClassDirectories(base.ConstantPathList):
    """A list of local directories, from which all `*.py` files will be considered action definitions.
    Each loaded definition is named after the filename stem and must contain an `Action` class.
    e.g. foo-bar.py may be referenced in a YAML workflow as `type: foo-bar`."""

    ENVIRONMENT_VARIABLE_NAME = "GRANA_ACTIONS_CLASS_DEFINITIONS_DIRECTORY"
    RC_PARAMETER_NAME = "action_classes_directories"
    WORKFLOW_CONFIG_PARAMETER_NAME = "action_classes_directories"


class ExternalPythonModulesPaths(base.ConstantPathList):
    """A list of local directories, which are added to the sys.path while loading any external modules."""

    ENVIRONMENT_VARIABLE_NAME = "GRANA_EXTERNAL_MODULES_PATHS"
    RC_PARAMETER_NAME = "external_python_modules_paths"
    WORKFLOW_CONFIG_PARAMETER_NAME = "external_python_modules_paths"


class SubprocessStreamBufferLimit(base.ConstantBase[int]):
    """An integer number defining the maximum line stdout/stderr length for a spawned subprocess."""

    ENVIRONMENT_VARIABLE_NAME = "GRANA_SUBPROCESS_STREAM_BUFFER_LIMIT"
    RC_PARAMETER_NAME = "subprocess_stream_buffer_limit"
    DEFAULT = 2**16  # asyncio.streams._DEFAULT_LIMIT

    def cast(self, value: t.Any) -> int:
        return int(value)


class TemplarModulesWhitelist(base.ConstantBase[dict[str, str]]):
    """List of modules allowed to be used in a template."""

    ENVIRONMENT_VARIABLE_NAME = "GRANA_TEMPLAR_MODULES_WHITELIST"
    RC_PARAMETER_NAME = "templar_modules_whitelist"
    WORKFLOW_CONFIG_PARAMETER_NAME = "templar_modules_whitelist"

    def _register_result(self, result: dict[str, str], source: base.ConstantSource) -> None:
        """Cumulative constant processing"""
        if self._result_and_source is not base.sentinel:
            prev_result = t.cast(dict[str, str], self._result_and_source[0])
            prev_result.update(result)  # type: ignore[union-attr]
            result = prev_result
            source = base.ConstantSource.MULTIPLE
        self._result_and_source = result, source

    def cast(self, value: t.Union[str, list[str], dict[str, str]]) -> dict[str, str]:
        if isinstance(value, dict):
            return value
        # Ensure that type is list[str]
        if isinstance(value, str):
            value = value.split(",")
        result: dict[str, str] = {}
        for item in value:
            item = item.strip()
            if ":" in item:
                k, v = item.split(":", 1)
                result[k] = v
            else:
                result[item] = item
        return result

    def default(self) -> dict[str, str]:
        """An empty dict"""
        return {}



"""Interactive dialog."""

import sys
import typing as t

try:
    import inquirer  # type: ignore
except ImportError:  # pragma: no cover
    inquirer = None

from ..exceptions import InteractionError

__all__ = [
    "run_dialog",
]


def run_dialog(choices: list[t.Tuple[str, str]], default: list[str]) -> list[str]:  # pragma: no cover
    """Invoke an inquirer dialog"""
    if inquirer is None:
        raise InteractionError("Inquirer is not installed. Try reinstalling grana with the `dialog` or `all` extras.")
    if not sys.stdin.isatty():
        raise InteractionError("Not a TTY")
    answers: dict[str, list[str]] = inquirer.prompt(
        questions=[
            inquirer.Checkbox(
                name="actions",
                message="Select actions (SPACE to check, RETURN to proceed)",
                choices=choices,
                default=default,
                carousel=True,
            )
        ],
        raise_keyboard_interrupt=True,
    )
    selected_action_names: list[str] = answers["actions"]
    return selected_action_names



"""
Common types.
"""

from .display.base import BaseDisplay
from .loader.base import AbstractBaseWorkflowLoader
from .strategy.base import BaseStrategy

LoaderClassType = type[AbstractBaseWorkflowLoader]
StrategyClassType = type[BaseStrategy]
DisplayClassType = type[BaseDisplay]
LoaderType = AbstractBaseWorkflowLoader
StrategyType = BaseStrategy
DisplayType = BaseDisplay

__all__ = [
    "LoaderClassType",
    "StrategyClassType",
    "DisplayClassType",
    "LoaderType",
    "StrategyType",
    "DisplayType",
]



"""DataClassLoader definition"""

import dataclasses
import itertools
import typing as t

from . import exceptions
from ...logging import WithLogger

try:
    from types import UnionType  # type: ignore[attr-defined]  # pylint: disable=no-name-in-module
except ImportError:  # pragma: no cover
    UnionType = None  # type: ignore[assignment, misc]

NoneType = type(None)
T = t.TypeVar("T")

__all__ = [
    "DataClassLoader",
]


class TypeSentinel:
    """Types sentinel"""


@dataclasses.dataclass
class DataClassLoader(WithLogger):
    """Constructs data classes by type and data"""

    validate_types: bool = True
    cast_types_list: list[type] = dataclasses.field(default_factory=list)
    NUMERIC_TYPES: t.ClassVar[list[type]] = [int, float, complex]

    def from_dict(self, data_type: type[T], data: t.Mapping) -> T:
        """Create an instance from a dict."""
        return self._internal_from_dict(data_type=data_type, data=data, path="")

    def _internal_from_dict(self, data_type: type[T], data: t.Mapping, path: str) -> T:
        # Get a list of init fields
        dataclass_init_fields: list[dataclasses.Field] = [
            f for f in dataclasses.fields(data_type) if f.init  # type: ignore[arg-type]
        ]
        # Check unexpected fields
        if unexpected_field_names := set(data.keys()) - {f.name for f in dataclass_init_fields}:
            raise exceptions.UnexpectedDataError(keys=unexpected_field_names, path=path)

        dataclass_type_hints = t.get_type_hints(data_type)
        init_values: t.MutableMapping[str, t.Any] = {}
        for field in dataclass_init_fields:
            field_path: str = f"{path}/{field.name}"
            if field.name not in data:
                # Check if default is set for the field
                if any(attr != dataclasses.MISSING for attr in (field.default, field.default_factory)):
                    continue
                raise exceptions.MissingValueError(field_path)
            field_type_hint = dataclass_type_hints[field.name]
            value = self._make_value_for_type(
                data_type=field_type_hint,
                data=data[field.name],
                path=field_path,
            )
            if self.validate_types and not self._is_instance(value, field_type_hint):
                raise exceptions.TypeMatchError(path=field_path, data_type=field_type_hint, value=value)
            init_values[field.name] = value

        return data_type(**init_values)

    def _make_value_for_type(self, data_type: type, data: t.Any, path: str) -> t.Any:
        if self._is_optional(data_type) and data is None:
            return data
        if self._is_union(data_type):
            data = self._make_value_for_union_type(union_data_type=data_type, data=data, path=path)
        elif self._extract_origin_from_generic_collection(data_type) is not TypeSentinel:
            data = self._make_value_for_collection_type(collection_data_type=data_type, data=data, path=path)
        elif dataclasses.is_dataclass(self._get_origin_for_dataclass(data_type)) and isinstance(data, t.Mapping):
            data = self._internal_from_dict(data_type=data_type, data=data, path=path)
        for cast_type in self.cast_types_list:
            if self._is_subclass(data_type, cast_type):
                return data_type(data)
        return data

    def _make_value_for_union_type(self, union_data_type: type, data: t.Any, path: str) -> t.Any:
        union_types = self._extract_generic_args(union_data_type)
        if self._is_optional(union_data_type) and len(union_types) == 2:
            for sub_type in union_types:
                if sub_type is not NoneType:
                    return self._make_value_for_type(data_type=sub_type, data=data, path=path)
            raise ValueError(f"Only None types found for union: {union_data_type!r}")
        union_matches = {}
        for inner_type in union_types:
            try:
                value = self._make_value_for_type(data_type=inner_type, data=data, path=path)
            except Exception as e:
                self.logger.debug(f"Non-successful union member type attempt: {e!r}")
                continue
            if self._is_instance(value, inner_type):
                union_matches[inner_type] = value
        if len(union_matches) > 1:
            raise exceptions.StrictUnionTypeMatchError(matches=union_matches, path=path)
        if union_matches:
            _, value = union_matches.popitem()
            return value
        if not self.validate_types:
            return data
        raise exceptions.UnionTypeMatchError(data_type=union_data_type, value=data, path=path)

    def _make_value_for_collection_type(self, collection_data_type: type, data: t.Any, path: str) -> t.Any:
        data_type = data.__class__
        if isinstance(data, t.Mapping) and self._is_subclass(collection_data_type, t.Mapping):
            if mapping_subscription := self._extract_generic_args(collection_data_type):
                _, value_type = mapping_subscription
            else:
                value_type = t.Any
            return data_type(
                (
                    key,
                    self._make_value_for_type(
                        data_type=value_type,
                        data=value,
                        path=f"{path}/{key}",
                    ),
                )
                for key, value in data.items()
            )
        if isinstance(data, tuple) and self._is_subclass(collection_data_type, tuple):
            # Check empty tuple
            if not data:
                return data_type()
            collection_types: tuple = self._extract_generic_args(collection_data_type)
            if len(collection_types) == 2 and collection_types[1] == Ellipsis:
                return data_type(
                    self._make_value_for_type(
                        data_type=collection_types[0],
                        data=item,
                        path=f"{path}@{num}",
                    )
                    for num, item in enumerate(data)
                )
            return data_type(
                self._make_value_for_type(
                    data_type=data_type,
                    data=item,
                    path=f"{path}@{num}",
                )
                for num, (item, data_type) in enumerate(itertools.zip_longest(data, collection_types))
            )
        if isinstance(data, t.Collection) and self._is_subclass(collection_data_type, t.Collection):
            value_type_hints: tuple = self._extract_generic_args(collection_data_type) or (t.Any,)
            return data_type(
                self._make_value_for_type(
                    data_type=value_type_hints[0],
                    data=item,
                    path=f"{path}@{num}",
                )
                for num, item in enumerate(data)
            )
        return data

    @classmethod
    def _get_origin_for_dataclass(cls, data_type: type) -> t.Any:
        if dataclasses.is_dataclass(data_type):
            return data_type
        return t.get_origin(data_type)

    @classmethod
    def _is_optional(cls, data_type: type) -> bool:
        return cls._is_union(data_type) and NoneType in cls._extract_generic_args(data_type)

    @classmethod
    def _is_union(cls, data_type: type) -> bool:
        # Check "A | B" form
        if UnionType is not None and isinstance(data_type, UnionType):
            return True
        return cls._extract_origin_from_generic(data_type) == t.Union

    @classmethod
    def _extract_origin_from_generic(cls, data_type: type) -> type:
        return getattr(data_type, "__origin__", TypeSentinel)

    @classmethod
    def _extract_origin_from_generic_collection(cls, data_type: type) -> type:
        if (origin := cls._extract_origin_from_generic(data_type)) is not TypeSentinel:
            try:
                if issubclass(origin, t.Collection):
                    return origin
            except (TypeError, AttributeError):
                pass
        return TypeSentinel

    @classmethod
    def _extract_generic_args(cls, data_type: type) -> tuple:
        if getattr(data_type, "_special", False):
            return ()
        return getattr(data_type, "__args__", ())

    @classmethod
    def _is_subclass(cls, sub_type: type, base_type: type) -> bool:
        if (origin := cls._extract_origin_from_generic_collection(sub_type)) is not TypeSentinel:
            sub_type = origin
        try:
            return issubclass(sub_type, base_type)
        except TypeError:
            return False

    # pylint: disable=too-many-return-statements
    @classmethod
    def _is_instance(cls, value: t.Any, data_type: type) -> bool:
        # typing.Any is always good
        if data_type == t.Any:
            return True
        # Check built-in isinstance
        try:
            if isinstance(value, data_type):
                return True
        except TypeError:
            # E.g. when data_type is not a type
            pass
        # Check numbers
        if data_type in cls.NUMERIC_TYPES and (value_type := type(value)) in cls.NUMERIC_TYPES:
            return cls.NUMERIC_TYPES.index(value_type) <= cls.NUMERIC_TYPES.index(data_type)
        # Check unions
        if cls._is_union(data_type):
            return any(cls._is_instance(value, tp) for tp in cls._extract_generic_args(data_type))
        # Check generic collections
        if (origin := cls._extract_origin_from_generic_collection(data_type)) is not TypeSentinel:
            # Upper-level check first
            if not isinstance(value, origin):
                return False
            if not cls._extract_generic_args(data_type):
                return True
            # Tuple case
            if isinstance(value, tuple) and cls._is_subclass(data_type, tuple):
                tuple_types = cls._extract_generic_args(data_type)
                # Empty tuple
                if len(tuple_types) == 1 and tuple_types[0] == ():
                    return len(value) == 0
                # Tuple with `...`
                if len(tuple_types) == 2 and tuple_types[1] is Ellipsis:
                    return all(cls._is_instance(item, tuple_types[0]) for item in value)
                if len(tuple_types) != len(value):
                    return False
                return all(cls._is_instance(item, item_type) for item, item_type in zip(value, tuple_types))
            # Mapping case
            if isinstance(value, t.Mapping):
                key_type, val_type = cls._extract_generic_args(data_type) or (t.Any, t.Any)
                for key, val in value.items():
                    if not cls._is_instance(key, key_type) or not cls._is_instance(val, val_type):
                        return False
                return True
            # Other iterables
            if not isinstance(value, t.Iterable):
                return False
            value_type_hints: tuple = cls._extract_generic_args(data_type) or (t.Any,)
            return all(cls._is_instance(item, value_type_hints[0]) for item in value)
        # Check for generic over type (e.g. type[MyClass])
        if cls._extract_origin_from_generic(data_type) in (type, t.Type):
            type_subscription: type = cls._extract_generic_args(data_type)[0]
            return cls._is_subclass(value, type_subscription)
        return False



"""YAML-based workflow load routines"""

from __future__ import annotations

import functools
import typing as t
from pathlib import Path

import yaml

from .base import AbstractBaseWorkflowLoader
from .utils import DefaultYAMLLoader
from ..actions.base import WorkflowActionExecution, ActionBase
from ..actions.bundled import (
    EchoAction,
    ShellAction,
    SubflowAction,
    LoopAction,
    DockerShellAction,
)
from ..config.constants import C
from ..config.constants.cache import CACHE
from ..config.constants.helpers import class_from_module

__all__ = [
    "DefaultYAMLWorkflowLoader",
]


class DefaultYAMLWorkflowLoader(AbstractBaseWorkflowLoader):
    """Default loader for YAML source files"""

    ALLOWED_ROOT_TAGS: set[str] = {"actions", "context", "miscellaneous", "configuration"}

    @classmethod
    def get_action_factories_info(cls) -> dict[str, tuple[type[ActionBase], str]]:
        return {
            **cls._get_static_action_factories_mapping(),
            **cls._load_external_action_factories_mapping(),
        }

    @classmethod
    @CACHE.wrap
    def _get_static_action_factories_mapping(cls) -> dict[str, tuple[type[ActionBase], str]]:
        return {
            name: (klass, "built-in")
            for name, klass in (
                ("echo", EchoAction),
                ("shell", ShellAction),
                ("subflow", SubflowAction),
                ("loop", LoopAction),
                ("docker-shell", DockerShellAction),
            )
            if klass is not None
        }

    @classmethod
    @CACHE.wrap
    def _load_external_action_factories_mapping(cls) -> dict[str, tuple[type[ActionBase], str]]:
        return cls._get_action_factories_from_sources_tuple(sources=tuple(C.ACTION_CLASSES_DIRECTORIES))

    @classmethod
    @functools.lru_cache(1)
    def _get_action_factories_from_sources_tuple(cls, sources: tuple[Path]) -> dict[str, tuple[type[ActionBase], str]]:
        dynamic_bases_map: dict[str, tuple[type[ActionBase], str]] = {}
        for class_directory_path in sources:  # type: Path
            class_directory_path = class_directory_path.resolve()
            if not class_directory_path.exists():
                cls.logger.warning(f"Given actions classes directory does not exist: {class_directory_path!r}")
                continue
            if not class_directory_path.is_dir():
                cls.logger.warning(f"Given actions classes path is not a directory: {class_directory_path!r}")
                continue
            cls.logger.info(f"Loading external action classes from {str(class_directory_path)!r}")
            for class_file in class_directory_path.iterdir():
                if not class_file.is_file() or not class_file.suffix == ".py":
                    continue
                action_type: str = class_file.stem
                cls.logger.debug(f"Trying external action class source: {class_file}")
                action_class: type[ActionBase] = t.cast(
                    type[ActionBase],
                    class_from_module(
                        source_path=class_file,
                        class_name="Action",
                        submodule_name=f"actions.{action_type}",
                    ),
                )
                if action_type in dynamic_bases_map:
                    cls.logger.warning(f"Class {action_type!r} is already defined: overriding from {class_file}")
                dynamic_bases_map[action_type] = (action_class, str(class_file))
        return dynamic_bases_map

    def _internal_load_from_text(self, data: str) -> None:
        root_node: dict = yaml.load(data, DefaultYAMLLoader)  # nosec
        if not isinstance(root_node, dict):
            self._throw(f"Unknown workflow structure: {type(root_node)!r} (should be a dict)")
        self._internal_load_from_dict(data=root_node)

    def _internal_load_from_dict(self, data: dict) -> None:
        root_keys: set[str] = set(data)
        if not root_keys:
            self._throw(f"Empty root dictionary (expected some of: {', '.join(sorted(self.ALLOWED_ROOT_TAGS))}")
        if unrecognized_keys := root_keys - self.ALLOWED_ROOT_TAGS:
            self._throw(
                f"Unrecognized root keys: {sorted(unrecognized_keys)} "
                f"(expected some of: {', '.join(sorted(self.ALLOWED_ROOT_TAGS))}"
            )
        processable_keys: set[str] = set(data) & self.ALLOWED_ROOT_TAGS
        if "configuration" in processable_keys:
            self.load_configuration_from_dict(data["configuration"])
        with self._loaded_config.apply():
            if "actions" in processable_keys:
                actions: list[dict] = data["actions"]
                if not isinstance(actions, list):
                    self._throw(f"'actions' contents should be a list (got {type(actions)!r})")
                for child_node in actions:
                    if not isinstance(child_node, dict):
                        self._throw(f"Unrecognized node type: {type(child_node)!r}")
                    action: WorkflowActionExecution = self.build_action_from_dict_data(child_node)
                    self._register_action(action)
            if "context" in processable_keys:
                context: dict[str, t.Any] = data["context"]
                if not isinstance(context, dict):
                    self._throw(f"'context' contents should be a dict (got {type(context)!r})")
                self._loads_contexts_dict(data=context)

    def _loads_contexts_dict(self, data: dict[str, t.Any]) -> None:
        for context_key, context_value in data.items():
            if not isinstance(context_key, str):
                self._throw(f"Context keys should be strings (got {type(context_key)!r} for {context_key!r})")
            self.logger.debug(f"Context key defined: {context_key}")
            self._gathered_context[context_key] = context_value



"""Lazy-loaded constants"""

import contextlib
import typing as t

from . import base, impl, rc, cache
from .cache import CACHE
from .cli import get_cli_option
from .helpers import class_from_module

__all__ = [
    "C",
]


class C:
    """Runtime constants"""

    LOG_LEVEL: base.ConstantBase = impl.LogLevel()
    LOG_FILE: base.ConstantBase = impl.LogFile()
    RC_FILE: base.ConstantBase = impl.RcFile()
    CONTEXT_DIRECTORY: base.ConstantBase = impl.ContextDirectory()
    RUNNER_TEMPORARY_DIRECTORY: base.ConstantBase = impl.RunnerTemporaryDirectory()
    INTERACTIVE_MODE: base.ConstantBase = impl.InteractiveMode()
    DEPENDENCY_DEFAULT_STRICTNESS: base.ConstantBase = impl.DependencyDefaultStrictness()
    WORKFLOW_SOURCE_FILE: base.ConstantBase = impl.WorkflowSourceFile()
    WORKFLOW_LOADER_CLASS: base.ConstantBase = impl.WorkflowLoaderClass()
    INTERNAL_DISPLAY_CLASS: base.ConstantBase = impl.InternalDisplayClass()
    EXTERNAL_DISPLAY_CLASS: base.ConstantBase = impl.ExternalDisplayClass()
    STRATEGY_CLASS: base.ConstantBase = impl.StrategyClass()
    USE_COLOR: base.ConstantBase = impl.UseColor()
    DEFAULT_SHELL_EXECUTABLE: base.ConstantBase = impl.DefaultShellExecutable()
    SHELL_INJECT_YIELD_FUNCTION: base.ConstantBase = impl.ShellInjectYieldFunction()
    STRICT_OUTCOMES_RENDERING: base.ConstantBase = impl.StrictOutcomesRendering()
    ACTION_CLASSES_DIRECTORIES: base.ConstantBase = impl.ActionClassDirectories()
    EXTERNAL_PYTHON_MODULES_PATHS: base.ConstantBase = impl.ExternalPythonModulesPaths()
    SUBPROCESS_STREAM_BUFFER_LIMIT: base.ConstantBase = impl.SubprocessStreamBufferLimit()
    TEMPLAR_MODULES_WHITELIST: base.ConstantBase = impl.TemplarModulesWhitelist()

    @classmethod
    def constants_info(cls) -> t.Iterable[base.ConstantProxyDescriptor]:
        """Return set of info for all constants"""
        for attr_name in sorted(dir(cls)):
            attr_value: t.Any = getattr(cls, attr_name)
            attr_value_origin: t.Any = getattr(attr_value, "__factory__", None)
            if isinstance(attr_value_origin, base.ConstantProxyDescriptor):
                yield attr_value_origin

    @classmethod
    def env_doc(cls) -> str:
        """Returns the info on the environment variables usage"""
        lines: list[str] = []
        for const_descriptor in cls.constants_info():
            lines.append(f"{const_descriptor.name}:")
            lines.extend(f"    {line.lstrip()}" for line in const_descriptor.definition.__doc__.splitlines())
        return "\n".join(lines)

    @classmethod
    @contextlib.contextmanager
    def mount_context_cache(cls) -> t.Generator[None, None, None]:
        """Enable context cache for all constants"""
        with CACHE.mount():
            yield

    @classmethod
    def reset_context_cache(cls) -> None:
        """Reset context cache for all constants"""
        CACHE.clear()



"""Deferred proxy helpers."""

import dataclasses
import typing as t

__all__ = [
    "DeferredCallsProxy",
]


@dataclasses.dataclass
class DeferredCall:
    """Represents a call that will happen later"""

    method: t.Callable
    args: t.Tuple = tuple()
    kwargs: dict[str, t.Any] = dataclasses.field(default_factory=dict)

    def __call__(self, *args, **kwargs) -> None:
        self.args = args
        self.kwargs = kwargs


class DeferredCallsProxy:
    """Stores all calls to the proxy object methods to evaluate them after calling the `uncork` method"""

    def __init__(self, obj: t.Any) -> None:
        self.__obj: t.Any = obj
        self.__deferred_calls: t.Optional[list[DeferredCall]] = []

    def uncork(self) -> None:
        """Make all deferred calls happen in the sequential manner"""
        if self.__deferred_calls is None:
            return
        for deferred_call in self.__deferred_calls:
            deferred_call.method(*deferred_call.args, **deferred_call.kwargs)
        self.__deferred_calls = None

    def __getattr__(self, item: str) -> t.Any:
        proxy_attr: t.Any = getattr(self.__obj, item)
        if self.__deferred_calls is None or not callable(proxy_attr):
            return proxy_attr
        self.__deferred_calls.append(result := DeferredCall(method=proxy_attr))
        return result



"""It's all about logging."""

import functools
import logging
import logging.config
import pathlib
import sys
import typing as t

from .tools.context import ContextManagerVar

__all__ = [
    "WithLogger",
    "context",
    "configure_logging",
]


class LoggerProperty:
    """Class-level logger property"""

    @staticmethod
    @functools.cache
    def _prepare(caller_type: type) -> logging.Logger:
        return logging.getLogger(f"{caller_type.__module__}.{caller_type.__name__}")

    def __get__(self, caller_instance: t.Any, caller_type: type) -> logging.Logger:
        return self._prepare(caller_type)


class WithLogger:
    """Add logger property"""

    logger = LoggerProperty()


COLOR_CODE_MAP: dict[str, int] = {
    "CRITICAL": 31,
    "FATAL": 31,
    "ERROR": 31,
    "WARN": 35,
    "WARNING": 35,
    "INFO": 34,
    "DEBUG": 32,
    "NOTSET": 37,
    "TRACE": 33,
}
DEFAULT_COLOR_CODE: int = 37

LOG_CONTEXT_DATA: ContextManagerVar[dict[str, t.Any]] = ContextManagerVar(default={})


class ContextFilter(logging.Filter):
    """Try filters"""

    def filter(self, record: logging.LogRecord) -> bool:
        if not (ctx := LOG_CONTEXT_DATA.get()):
            record.__dict__["context"] = ""
            return True
        ctx_pairs = (f"{k}={ctx[k]}" for k in sorted(ctx))
        record.__dict__["context"] = f" {{{', '.join(ctx_pairs)}}}"
        return True


def context(**kwargs):
    """Manager for log record context data"""
    return LOG_CONTEXT_DATA.set(kwargs)


class MonochromeFormatter(logging.Formatter):
    """No colors"""

    def __init__(self):
        super().__init__(fmt="{asctime} {levelname} [{name}]{context} {message}", style="{")


class ColorFormatter(logging.Formatter):
    """With colors in the level name"""

    def __init__(self):
        super().__init__(fmt="{asctime} {colored_levelname} [{name}]{colored_context} {message}", style="{")

    def format(self, record: logging.LogRecord) -> str:
        record.__dict__["colored_levelname"] = self._colorize_level_name(record.levelname)
        colored_context: str = ""
        if ctx := record.__dict__.get("context", ""):
            colored_context = self._colorize_string_by_code(ctx, 37)  # Gray
        record.__dict__["colored_context"] = colored_context
        return super().format(record)

    @classmethod
    @functools.cache
    def _colorize_level_name(cls, name: str) -> str:
        code: int = COLOR_CODE_MAP.get(name, DEFAULT_COLOR_CODE)
        return cls._colorize_string_by_code(name, code)

    @classmethod
    def _colorize_string_by_code(cls, string: str, code: int) -> str:
        return f"\033[{code}m{string}\033[0m"


def get_main_logger() -> logging.Logger:
    """Return the root logger for the package"""
    # Trim subpackage name
    root_logger_name: str = __name__.removesuffix(".logging")
    return logging.getLogger(root_logger_name)


def configure_logging(level: str, colorize: bool = False, main_file: t.Optional[pathlib.Path] = None) -> None:
    """Logging setup"""
    main_logger = get_main_logger()
    main_logger.setLevel(level)
    ctx_filter = ContextFilter()

    if not main_file:
        # Process stdout handler
        stderr_handler = logging.StreamHandler(sys.stderr)
        formatter: logging.Formatter = ColorFormatter() if colorize else MonochromeFormatter()
        stderr_handler.setFormatter(formatter)
        stderr_handler.addFilter(ctx_filter)
        main_logger.addHandler(stderr_handler)
    else:
        # Process file handler
        filename = main_file.expanduser().resolve()
        # Prepare parent directory
        filename.parent.mkdir(parents=True, exist_ok=True)
        file_handler = logging.FileHandler(filename, mode="w")
        file_handler.setFormatter(MonochromeFormatter())
        file_handler.addFilter(ctx_filter)
        main_logger.addHandler(file_handler)



"""Display utility functions."""

import dataclasses
import itertools
import typing as t

__all__ = [
    "Tree",
    "locate_parent_name_by_prefix",
]

T = t.TypeVar("T")
AsciiLinesGeneratorType = t.Generator[t.Tuple[str, T], None, None]


@dataclasses.dataclass
class Node(t.Generic[T]):
    """Topology relations representation"""

    content: T
    children: list


class Tree(dict[str, Node[T]]):
    """Generic tree builder"""

    def __init__(self):
        super().__init__()
        self.__root_nodes_list: list[Node[T]] = []

    def put(self, items: t.Iterable[t.Tuple[str, T]], *, parent_name: t.Optional[str] = None) -> None:
        """Put items into the topology"""
        nodes_list: list[Node[T]] = []
        for name, item in items:
            node: Node[T] = Node(content=item, children=[])
            nodes_list.append(node)
            self[name] = node
        if parent_name is None:
            self.__root_nodes_list.extend(nodes_list)
        else:
            self[parent_name].children.extend(nodes_list)

    def generate_ascii_representation(self) -> AsciiLinesGeneratorType:
        """Generate tree representation components. A component is a pair of:
        - Prefix: a string containing aligned ASCII box drawing symbols, respective to the tree node graph
        - Object: the content of the corresponding node
        """
        yield from self._internal_tree_generate(nodes=self.__root_nodes_list, prefix=None)

    def _internal_tree_generate(self, nodes: list[Node[T]], prefix: t.Optional[str]) -> AsciiLinesGeneratorType:
        last_node_num: int = len(nodes) - 1
        for num, node in enumerate(nodes):
            is_last_node: bool = num == last_node_num
            if prefix is None:
                yield "", node.content
                yield from self._internal_tree_generate(
                    nodes=node.children,
                    prefix="",
                )
            else:
                yield prefix + ("└──" if is_last_node else "├──"), node.content
                yield from self._internal_tree_generate(
                    nodes=node.children,
                    prefix=prefix + ("   " if is_last_node else "│  "),
                )


def get_common_prefix(*strings: str) -> str:
    """Calculate the longest common prefix of a sequence of strings"""
    character_tuples: t.Iterable[t.Tuple[str, ...]] = zip(*strings)
    common_prefix_iterator = itertools.takewhile(lambda chars: all(chars[0] == c for c in chars), character_tuples)
    return "".join(common_chars[0] for common_chars in common_prefix_iterator)


def locate_parent_name_by_prefix(children: t.Iterable[str], candidates: t.Iterable[str]) -> str:
    """Define the optimal parent among candidates for the children"""
    longest_match_length: int = -1
    sources_common_prefix: str = get_common_prefix(*children)
    optimal_item_name: str = ""
    for candidate in candidates:
        match_length = len(get_common_prefix(candidate, sources_common_prefix))
        if match_length > longest_match_length:
            longest_match_length = match_length
            optimal_item_name = candidate
        elif match_length == longest_match_length and len(optimal_item_name) > len(candidate):
            optimal_item_name = candidate
    return optimal_item_name



"""
A strategy is an async-iterable object,
emitting actions one by one for further scheduling.
This module is for abstract base only.
"""

from __future__ import annotations

import typing as t

from ..actions.base import WorkflowActionExecution
from ..logging import WithLogger
from ..workflow import Workflow

ST = t.TypeVar("ST", bound="BaseStrategy")

__all__ = [
    "BaseStrategy",
    "STRATEGIES_MAP",
]

STRATEGIES_MAP: dict[str, type[BaseStrategy]] = {}


class BaseStrategy(WithLogger, t.AsyncIterable[WorkflowActionExecution]):
    """Strategy abstract base"""

    NAME: str = ""

    @staticmethod
    def get_strategy_class_by_name(name: str) -> type[BaseStrategy]:
        """Obtain derived strategy class by name"""
        try:
            return STRATEGIES_MAP[name]
        except KeyError:
            raise ValueError(f"Invalid strategy name: {name!r} (allowed: {sorted(STRATEGIES_MAP)})") from None

    def __init__(self, workflow: Workflow) -> None:
        self._workflow = workflow

    def __aiter__(self: ST) -> ST:
        return self

    async def __anext__(self) -> WorkflowActionExecution:
        raise NotImplementedError

    def __init_subclass__(cls, **kwargs):
        if STRATEGIES_MAP.setdefault(cls.NAME, cls) is not cls:
            raise NameError(
                f"Strategy named {cls.NAME!r} already exists. "
                f"Please specify another name for the {cls.__module__}.{cls.__name__}."
            )



"""Templar containers for rendering."""

import importlib
import types
import typing as t

from .proxy import LazyProxy
from ..exceptions import ActionRenderError, PendingActionUnresolvedOutcomeError
from ..logging import WithLogger

__all__ = [
    "AttrDict",
    "LooseDict",
    "OutcomeDict",
    "ActionContainingDict",
    "ExternalModulesDict",
    "LazyProxy",
]

RenderHookType = t.Callable[[str], str]


class ItemAttributeAccessorMixin:
    """Anything, that can be accessed fie __getitem__, is available also as an attribute"""

    def __getattr__(self, item: str):
        return self.__getitem__(item)


class AttrDict(dict, ItemAttributeAccessorMixin):
    """Basic dictionary that allows attribute read access to its keys"""


class LooseDict(AttrDict):
    """A dictionary that allows attribute read access to its keys with a default empty value fallback"""

    def __getitem__(self, item: str):
        try:
            return super().__getitem__(item)
        except KeyError:
            return ""


class OutcomeDict(AttrDict):
    """A dictionary that allows attribute read access to its keys with a default value fallback"""

    def __getitem__(self, item: str):
        try:
            return super().__getitem__(item)
        except KeyError as e:
            from ..config.constants import C

            if C.STRICT_OUTCOMES_RENDERING:
                raise ActionRenderError(f"Outcome key {e} not found") from e
            return ""


class ActionContainingDict(AttrDict):
    """Anything with action names as keys"""

    def __getitem__(self, item: str):
        try:
            return super().__getitem__(item)
        except KeyError as e:
            raise ActionRenderError(f"Action not found: {e}") from e


class ActionOutcomeAggregateDict(ActionContainingDict):
    """Top-level container for action outcomes"""

    def __getitem__(self, item: str):
        if (result := super().__getitem__(item)) is not None:
            return result
        raise PendingActionUnresolvedOutcomeError(item)


class ExternalModulesDict(AttrDict, WithLogger):
    """Container for accessing external modules inside templars"""

    def __init__(self):
        super().__init__()
        self._known_modules: t.Dict[str, types.ModuleType] = {}

    def __getitem__(self, key):
        if key not in self._known_modules:
            self._known_modules[key] = self._load_module(key)
        return self._known_modules[key]

    @classmethod
    def _load_module(cls, module_alias: str):
        from ..config.constants import C, helpers

        if (actual_module_name := C.TEMPLAR_MODULES_WHITELIST.get(module_alias)) is None:
            raise KeyError(f"Module alias {module_alias!r} is not added to the templar modules whitelist")
        cls.logger.debug(f"Module alias {module_alias!r} resolved to {actual_module_name!r}")
        with helpers.add_sys_paths(*C.EXTERNAL_PYTHON_MODULES_PATHS):
            return importlib.import_module(actual_module_name)



"""Loader context"""

import contextlib
import typing as t
from pathlib import Path

from ..rendering import CommonTemplar
from ..tools.context import ContextManagerVar

__all__ = [
    "LOADED_FILE_STACK",
]


class LoadedFileStack:
    """Context variable that points to the loading file,
    which may be either a workflow file, runtime configuration file
    or a simple YAML file loaded by the `!load` tag."""

    def __init__(self):
        self._stack: ContextManagerVar[tuple[Path, ...]] = ContextManagerVar(default=tuple())

    @contextlib.contextmanager
    def add(self, path: Path) -> t.Generator[None, None, None]:
        """Add a path to the stack"""

        current_stack: tuple[Path, ...] = self._stack.get()
        with self._stack.set((path,) + current_stack):
            yield

    def get_all(self) -> tuple[Path, ...]:
        """Return all loading files in reversed order (from the active to the first)"""
        return self._stack.get()

    def get_last(self) -> t.Optional[Path]:
        """Return loading file, if any"""
        if not (stack := self.get_all()):
            return None
        return stack[0]

    def create_associated_templar(self) -> CommonTemplar:
        """Create proper common templar based on the current context."""
        current_loaded_file: t.Optional[Path] = self.get_last()
        if current_loaded_file is not None:
            return CommonTemplar.from_source_file(path=current_loaded_file)
        # This happens when the runner is fed from stdin
        return CommonTemplar.from_context_directory()


LOADED_FILE_STACK = LoadedFileStack()



"""Workflow-related module"""

from __future__ import annotations

import collections
import pathlib
import typing as t

from .actions.base import WorkflowActionExecution, ActionDependency
from .exceptions import IntegrityError
from .logging import WithLogger
from .rendering import WorkflowTemplar
from .config.constants.workflow import WorkflowConfiguration

__all__ = [
    "Workflow",
]


class Workflow(dict[str, WorkflowActionExecution], WithLogger):
    """Action relations map"""

    def __init__(
        self,
        actions_map: dict[str, WorkflowActionExecution],
        context: t.Optional[dict[str, t.Any]] = None,
        source_file: t.Optional[pathlib.Path] = None,
        configuration: t.Optional[WorkflowConfiguration] = None,
    ) -> None:
        super().__init__(actions_map)
        self.source_file: t.Optional[pathlib.Path] = source_file
        self.configuration: WorkflowConfiguration = configuration or WorkflowConfiguration()
        self._entrypoints: set[str] = set()
        self._tiers_sequence: list[list[WorkflowActionExecution]] = []
        self._descendants_map: dict[str, dict[str, ActionDependency]] = collections.defaultdict(dict)
        self.context: dict[str, t.Any] = context or {}
        # Check dependencies integrity
        self._establish_descendants()
        # Create order map to check all actions are reachable
        self._allocate_tiers()

    def get_templar(self, locals_map: dict) -> WorkflowTemplar:
        """Create a Templar object"""
        return WorkflowTemplar(
            outcomes_map={name: self[name].outcomes for name in self},
            action_states={name: self[name].status.value for name in self},
            context_map=self.context,
            metadata=self.get_metadata(),
            locals_map=locals_map,
        )

    def get_metadata(self) -> dict[str, t.Any]:
        """Obtain workflow metadata for further use in templating"""
        from .config.constants import C

        metadata: dict[str, t.Any] = {
            "cwd": C.CONTEXT_DIRECTORY,
            "tempdir": C.RUNNER_TEMPORARY_DIRECTORY,
        }
        if self.source_file is not None:
            source_file: pathlib.Path = self.source_file.resolve()
            metadata.update(
                {
                    "source_file": source_file,
                    "here": source_file.parent,
                }
            )
        return metadata

    def _establish_descendants(self) -> None:
        missing_deps: set[str] = set()
        for action_execution in self.values():  # type: WorkflowActionExecution
            for dependency in action_execution.ancestors:
                if dependency.name not in self:
                    missing_deps.add(dependency.name)
                    continue
                # Register symmetric descendant connection for further simplicity
                self._descendants_map[dependency.name][action_execution.name] = dependency
            # Check if there are any dependencies after removal at all
            if not action_execution.ancestors:
                self._entrypoints.add(action_execution.name)
        if missing_deps:
            raise IntegrityError(f"Missing actions among dependencies: {sorted(missing_deps)}")
        # Check entrypoints presence
        if not self._entrypoints:
            raise IntegrityError("No entrypoints for the workflow")

    def _allocate_tiers(self) -> None:
        """Use Dijkstra algorithm to introduce partial order.
        The tier is a group of tasks, and it's called finished when all its tasks are processed.
        Tier #0 consists of entrypoints.
        Tier #N consists of all tasks requiring exactly N-1 preceding tiers to be finished.
        """
        step_tier: int = 0
        action_execution_name_to_tier_mapping: dict[str, int] = {}
        #
        current_tier_actions_executions_names: set[str] = self._entrypoints
        while True:
            next_tier_candidate_actions_executions_names: set[str] = set()
            for tier_action_execution_name in current_tier_actions_executions_names:
                tier_action_execution: WorkflowActionExecution = self[tier_action_execution_name]
                if tier_action_execution.name in action_execution_name_to_tier_mapping:
                    continue
                action_execution_name_to_tier_mapping[tier_action_execution.name] = step_tier
                next_tier_candidate_actions_executions_names |= set(self._descendants_map[tier_action_execution_name])
            if not next_tier_candidate_actions_executions_names:
                break
            step_tier += 1
            current_tier_actions_executions_names = next_tier_candidate_actions_executions_names
        self.logger.debug(f"Number of tiers: {step_tier + 1}")
        unreachable_action_executions_names: set[str] = {
            execution.name for execution in self.values() if execution.name not in action_execution_name_to_tier_mapping
        }
        if unreachable_action_executions_names:
            raise IntegrityError(f"Unreachable actions found: {sorted(unreachable_action_executions_names)}")
        self._tiers_sequence = [[] for _ in range(step_tier + 1)]
        for action_execution_name, action_execution in self.items():
            action_execution_tier: int = action_execution_name_to_tier_mapping[action_execution_name]
            self._tiers_sequence[action_execution_tier].append(action_execution)

    def _iter_action_executions_by_tier(self) -> t.Generator[t.Tuple[int, WorkflowActionExecution], None, None]:
        """Yield actions tier by tier"""
        for tier_num, tier_executions in enumerate(self._tiers_sequence):
            for execution in tier_executions:
                yield tier_num, execution

    def iterate_actions(self) -> t.Iterator[WorkflowActionExecution]:
        """Iterate action executions sorted natively"""
        return (execution for _, execution in self._iter_action_executions_by_tier())



"""Runner-related constants helpers"""

import typing as t
from pathlib import Path

from ...tools.context import ContextManagerVar

__all__ = [
    "TEMP_DIR_CONTEXT",
]

TEMP_DIR_CONTEXT: ContextManagerVar[t.Optional[Path]] = ContextManagerVar(default=None)



"""Context variables base"""

import contextlib
import contextvars
import functools
import typing as t
import uuid

__all__ = [
    "ContextManagerVar",
    "ContextCache",
]

VT = t.TypeVar("VT")


class ContextManagerVar(t.Generic[VT]):
    """ContextVar with context manager"""

    def __init__(self, **kwargs):
        self._ctx_var: contextvars.ContextVar[VT] = contextvars.ContextVar(uuid.uuid4().hex, **kwargs)

    @contextlib.contextmanager
    def set(self, value):
        """Set context value"""
        token = self._ctx_var.set(value)
        try:
            yield
        finally:
            self._ctx_var.reset(token)

    def get(self) -> VT:
        """Get context value"""
        return self._ctx_var.get()


class ContextCache:
    """Context-based cache for functions"""

    def __init__(self):
        self._cache: ContextManagerVar[dict] = ContextManagerVar()

    @contextlib.contextmanager
    def mount(self) -> t.Generator[None, None, None]:
        """Starts a new context for the cache"""
        with self._cache.set({}):
            yield

    def wrap(self, f: t.Callable) -> t.Callable:
        """A decorator to apply the cache to the function"""

        @functools.wraps(f)
        def wrapped(*args, **kwargs):
            try:
                local_context_cache_dict: dict = self._cache.get()
            except LookupError:
                # Cache is not mounted
                return f(*args, **kwargs)
            # Prepare the signature to store result
            key = functools._make_key((f,) + args, kwargs, False)  # pylint: disable=protected-access
            if key not in local_context_cache_dict:
                local_context_cache_dict[key] = f(*args, **kwargs)
            return local_context_cache_dict[key]

        return wrapped

    def clear(self) -> None:
        """Clears the cache for the active context"""
        self._cache.get().clear()



"""Lazy proxy implementation"""

import operator
import os
import typing as t
import weakref
from functools import cached_property

__all__ = [
    "LazyProxy",
]


def _make_method(func: t.Callable) -> t.Callable:
    """Create proxy method"""

    def method_over_wrapped(self, *args, **kwargs):
        return func(self.__wrapped__, *args, **kwargs)

    return method_over_wrapped


def _make_r_method(func: t.Callable) -> t.Callable:
    """Create proxy method with positional args in reversed order"""

    def method_over_wrapped(self, other):
        return func(other, self.__wrapped__)

    return method_over_wrapped


# pylint: disable=inconsistent-return-statements
class LazyProxy:
    """Perform evaluation only once when it is required first time"""

    def __init__(self, factory):
        self.__dict__["__factory__"] = factory

    @cached_property
    def __wrapped__(self):
        return self.__dict__["__factory__"]()

    def __setattr__(self, __name, __value) -> None:
        if __name not in ("__wrapped__", "__factory__"):
            return setattr(self.__wrapped__, __name, __value)
        self.__dict__[__name] = __value

    def __getattr__(self, __name) -> t.Any:
        if __name not in ("__wrapped__", "__factory__"):
            return getattr(self.__wrapped__, __name)
        return self.__dict__[__name]

    def __delattr__(self, __name):
        if __name not in ("__wrapped__", "__factory__"):
            return delattr(self.__wrapped__, __name)
        del self.__dict__[__name]

    # operator.call is available only since python 3.11
    def __call__(self, *args, **kwargs):
        return self.__wrapped__(*args, **kwargs)

    __name__ = property(_make_method(operator.attrgetter("__name__")))  # type: ignore[assignment]
    __module__ = property(_make_method(operator.attrgetter("__module__")))  # type: ignore[assignment]
    __doc__ = property(_make_method(operator.attrgetter("__doc__")))  # type: ignore[assignment]
    __annotations__ = property(_make_method(operator.attrgetter("__annotations__")))  # type: ignore[assignment]
    __class__ = property(_make_method(operator.attrgetter("__class__")))  # type: ignore[assignment]
    __weakref__ = property(_make_method(weakref.ref))
    __enter__ = _make_method(operator.attrgetter("__enter__"))
    __exit__ = _make_method(operator.attrgetter("__exit__"))
    __dir__ = _make_method(dir)
    __str__ = _make_method(str)
    __bytes__ = _make_method(bytes)
    __repr__ = _make_method(repr)
    __reversed__ = _make_method(reversed)
    __round__ = _make_method(round)
    __lt__ = _make_method(operator.lt)
    __le__ = _make_method(operator.le)
    __eq__ = _make_method(operator.eq)
    __ne__ = _make_method(operator.ne)
    __gt__ = _make_method(operator.gt)
    __ge__ = _make_method(operator.ge)
    __hash__ = _make_method(hash)
    __bool__ = _make_method(bool)
    __add__ = _make_method(operator.add)
    __sub__ = _make_method(operator.sub)
    __mul__ = _make_method(operator.mul)
    __matmul__ = _make_method(operator.matmul)
    __truediv__ = _make_method(operator.truediv)
    __floordiv__ = _make_method(operator.floordiv)
    __mod__ = _make_method(operator.mod)
    __divmod__ = _make_method(divmod)
    __pow__ = _make_method(pow)
    __lshift__ = _make_method(operator.lshift)
    __rshift__ = _make_method(operator.rshift)
    __and__ = _make_method(operator.and_)
    __xor__ = _make_method(operator.xor)
    __or__ = _make_method(operator.or_)
    __radd__ = _make_r_method(operator.add)
    __rsub__ = _make_r_method(operator.sub)
    __rmul__ = _make_r_method(operator.mul)
    __rmatmul__ = _make_r_method(operator.matmul)
    __rtruediv__ = _make_r_method(operator.truediv)
    __rfloordiv__ = _make_r_method(operator.floordiv)
    __rmod__ = _make_r_method(operator.mod)
    __rdivmod__ = _make_r_method(divmod)
    __rpow__ = _make_r_method(pow)
    __rlshift__ = _make_r_method(operator.lshift)
    __rrshift__ = _make_r_method(operator.rshift)
    __rand__ = _make_r_method(operator.and_)
    __rxor__ = _make_r_method(operator.xor)
    __ror__ = _make_r_method(operator.or_)
    __iadd__ = _make_method(operator.iadd)
    __isub__ = _make_method(operator.isub)
    __imul__ = _make_method(operator.imul)
    __imatmul__ = _make_method(operator.imatmul)
    __itruediv__ = _make_method(operator.itruediv)
    __ifloordiv__ = _make_method(operator.ifloordiv)
    __imod__ = _make_method(operator.imod)
    __ipow__ = _make_method(operator.ipow)
    __ilshift__ = _make_method(operator.ilshift)
    __irshift__ = _make_method(operator.irshift)
    __iand__ = _make_method(operator.iand)
    __ixor__ = _make_method(operator.ixor)
    __ior__ = _make_method(operator.ior)
    __neg__ = _make_method(operator.neg)
    __pos__ = _make_method(operator.pos)
    __abs__ = _make_method(operator.abs)
    __invert__ = _make_method(operator.invert)
    __int__ = _make_method(int)
    __float__ = _make_method(float)
    __oct__ = _make_method(oct)
    __hex__ = _make_method(hex)
    __index__ = _make_method(operator.index)
    __len__ = _make_method(len)
    __contains__ = _make_method(operator.contains)
    __getitem__ = _make_method(operator.getitem)
    __setitem__ = _make_method(operator.setitem)
    __delitem__ = _make_method(operator.delitem)
    __iter__ = _make_method(iter)
    __fspath__ = _make_method(os.fspath)



"""Runtime values concealment helpers."""

import typing as t

__all__ = [
    "represent_object_type",
]


def represent_object_type(obj: t.Any) -> str:
    """Build a string representation of an object type to disguise real values"""
    if isinstance(obj, t.Mapping):
        if not obj:
            return "dict"
        return f"dict[{_represent_collection_as_union(obj)}, {_represent_collection_as_union(obj.values())}]"
    if isinstance(obj, list):
        if not obj:
            return "list"
        return f"list[{_represent_collection_as_union(obj)}]"
    return obj.__class__.__name__


def _represent_collection_as_union(collection: t.Iterable) -> str:
    sorted_unique_types_names: list[str] = sorted({represent_object_type(item) for item in collection})
    if len(sorted_unique_types_names) == 1:
        return sorted_unique_types_names[0]
    none_type_name: str = type(None).__name__
    if len(sorted_unique_types_names) == 2 and none_type_name in sorted_unique_types_names:
        non_none_types = [v for v in sorted_unique_types_names if v != none_type_name]
        return f"typing.Optional[{non_none_types[0]}]"
    return f"typing.Union[{', '.join(sorted_unique_types_names)}]"



"""All bundled action classes"""

from .echo import EchoAction
from .shell import ShellAction
from .subflow import SubflowAction
from .loop import LoopAction

try:
    from .docker_shell import DockerShellAction
except ImportError:  # pragma: no cover
    DockerShellAction = None  # type: ignore



"""Common loader utilities"""

import io
import typing as t
from pathlib import Path

from .base import AbstractBaseWorkflowLoader
from .default import DefaultYAMLWorkflowLoader
from ..exceptions import SourceError

__all__ = [
    "get_default_loader_class_for_source",
]

STREAM_DEFAULT_LOADER: type[AbstractBaseWorkflowLoader] = DefaultYAMLWorkflowLoader
DICT_DEFAULT_LOADER: type[AbstractBaseWorkflowLoader] = DefaultYAMLWorkflowLoader
SUFFIX_TO_LOADER_MAP: dict[str, type[AbstractBaseWorkflowLoader]] = {
    ".yml": DefaultYAMLWorkflowLoader,
    ".yaml": DefaultYAMLWorkflowLoader,
}


def get_default_loader_class_for_source(
    source: t.Union[Path, dict, io.TextIOBase],
) -> type[AbstractBaseWorkflowLoader]:
    """Return loader class based on file stats"""
    if isinstance(source, io.TextIOBase):
        return STREAM_DEFAULT_LOADER
    if isinstance(source, dict):
        return DICT_DEFAULT_LOADER
    if (loader_class := SUFFIX_TO_LOADER_MAP.get(source.suffix)) is None:
        raise SourceError(f"Unrecognized source: {source}")
    return loader_class



# pylint: disable=invalid-field-call
"""Separate module for loop action"""

import dataclasses
import itertools
import typing as t

from .subflow import make_subflow_runner_class_for_action
from ..base import ArgsBase, ActionBase
from ...actions import constants
from ...exceptions import ExecutionFailed
from ...rendering import CommonTemplar, containers as c

__all__ = [
    "LoopAction",
]

TOP_LEVEL_ONLY_ACTION_RESERVED_FIELD_NAMES: set[str] = constants.ACTION_RESERVED_FIELD_NAMES - {"type", "name"}


class LoopArgs(ArgsBase):
    """Loop step arguments."""

    matrix: dict[str, t.Any]
    step: dict[str, t.Any] = dataclasses.field(metadata={"rendering": "disabled"})
    strategy: t.Optional[str] = None
    strict: t.Optional[bool] = None
    actions: list[dict] = dataclasses.field(default_factory=list, init=False)


class LoopAction(ActionBase):
    """Executes an independent workflow and passes the display events to the original runner."""

    args: LoopArgs

    def on_render(self) -> None:
        if unexpected_fields := set(self.args.step) & TOP_LEVEL_ONLY_ACTION_RESERVED_FIELD_NAMES:
            raise ValueError(f"Unexpected `step` fields: {sorted(unexpected_fields)}")
        # Prepare ranges
        var_names: list[str] = []
        all_var_values: list[t.Any] = []
        for var_name, var_values in self.args.matrix.items():
            var_names.append(var_name)
            all_var_values.append(var_values)
        # Build actions
        actions: list[dict] = []
        for var_values_tuple in itertools.product(*all_var_values):
            matrix_locals_map: dict[str, t.Any] = c.AttrDict(zip(var_names, var_values_tuple))
            templar: CommonTemplar = self._communicator.get_templar(extra_locals=matrix_locals_map)
            action_dict: dict = templar.render(self.args.step)
            actions.append(action_dict)
        # Finally store built actions
        self.args.actions = actions

    async def run(self) -> None:
        from ...config.constants import C

        loop_runner_class = make_subflow_runner_class_for_action(self)

        # Cache [re]mount is required since the subflow may reconfigure some fields
        with C.mount_context_cache():
            # Prepare configuration section
            configuration: dict[str, t.Any] = {}
            if self.args.strategy is not None:
                configuration["strategy"] = self.args.strategy
            if self.args.strict is not None:
                configuration["strict"] = self.args.strict
            runner = loop_runner_class({"actions": self.args.actions, "configuration": configuration})
            try:
                await runner.run_async()
            except ExecutionFailed:
                self.fail()



"""Workflow-based configuration values"""

import contextlib
import dataclasses
import pathlib
import typing as t

from ...tools.context import ContextManagerVar

__all__ = [
    "WorkflowConfiguration",
    "CONTEXT_HOLDER",
]

CONTEXT_HOLDER: ContextManagerVar[dict[str, t.Any]] = ContextManagerVar(default={})


class ConfigSentinel:
    """Sentinel type for workflow configuration parameters"""


sentinel = ConfigSentinel()


@dataclasses.dataclass
class WorkflowConfiguration:
    """Configuration loaded from the workflow"""

    strategy: t.Union[str, ConfigSentinel] = sentinel
    default_shell_executable: t.Union[str, ConfigSentinel] = sentinel
    shell_inject_yield_function: t.Union[bool, ConfigSentinel] = sentinel
    strict_outcomes_rendering: t.Union[bool, ConfigSentinel] = sentinel
    action_classes_directories: t.Union[list[pathlib.Path], ConfigSentinel] = sentinel
    external_python_modules_paths: t.Union[list[pathlib.Path], ConfigSentinel] = sentinel
    strict: t.Union[bool, ConfigSentinel] = sentinel
    templar_modules_whitelist: t.Union[list[str], ConfigSentinel] = sentinel

    @contextlib.contextmanager
    def apply(self) -> t.Generator[None, None, None]:
        """Apply contextual values for the workflow configuration"""
        cfg_dict: t.Dict[str, t.Any] = {
            k: v for k, v in dataclasses.asdict(self).items() if not isinstance(v, ConfigSentinel)
        }
        with CONTEXT_HOLDER.set(cfg_dict):
            yield



"""Package version"""

import importlib.metadata

# Define default for missing installation (e.g. development environment with cloned source code)
__version__: str = "0.0.0"
try:
    __version__ = importlib.metadata.version(__package__)
except importlib.metadata.PackageNotFoundError:  # pragma: no cover
    pass

__all__ = [
    "__version__",
]



"""All the templating stuff."""

from __future__ import annotations

import os
import pathlib
import typing as t

from . import containers as c
from .constants import MAX_RECURSION_DEPTH
from .tokenizing import TemplarStringLexer
from ..actions.types import Expression, qualify_string_as_potentially_renderable, ActionStatus
from ..exceptions import (
    ActionRenderError,
    RestrictedBuiltinError,
    ActionRenderRecursionError,
    PendingActionUnresolvedOutcomeError,
)
from ..logging import WithLogger

__all__ = [
    "CommonTemplar",
    "WorkflowTemplar",
]


class CommonTemplar(WithLogger):
    """Expression renderer"""

    DISABLED_GLOBALS: list[str] = ["exec", "eval", "compile", "setattr", "delattr"]

    def __init__(self, **args: dict[str, t.Any]) -> None:
        self._globals: dict[str, t.Any] = {f: self._make_restricted_builtin_call_shim(f) for f in self.DISABLED_GLOBALS}
        self._globals.update(args)
        self._depth: int = 0

    @classmethod
    def _make_base_templar_with_meta_and_env(cls, extra_meta: t.Optional[dict[str, t.Any]] = None) -> CommonTemplar:
        from ..config.constants import C

        meta_dict: dict = c.LooseDict(cwd=C.CONTEXT_DIRECTORY)
        if extra_meta is not None:
            meta_dict.update(extra_meta)
        env_dict: dict = c.LooseDict(os.environ)
        return cls(
            metadata=meta_dict,
            environment=env_dict,
            # Aliases
            meta=meta_dict,
            env=env_dict,
        )

    @classmethod
    def from_source_file(cls, path: pathlib.Path) -> CommonTemplar:
        """Construct a base templar from the source file path"""
        resolved_path: pathlib.Path = path.resolve()
        return cls._make_base_templar_with_meta_and_env(
            extra_meta={
                "source_file": resolved_path,
                "here": resolved_path.parent,
            }
        )

    @classmethod
    def from_context_directory(cls) -> CommonTemplar:
        """Construct a base templar from the source file path"""
        from ..config.constants import C

        return cls._make_base_templar_with_meta_and_env(
            extra_meta={
                "here": C.CONTEXT_DIRECTORY,
            }
        )

    def _render_string(self, value: str) -> str:
        """Process string data, replacing all @{} occurrences."""
        try:
            return self._internal_render_string(value)
        except ActionRenderRecursionError as e:
            # Eliminate ActionRenderRecursionError stack trace on hit
            self.logger.debug(f"Rendering {value!r} failed: {e!r}")
            raise ActionRenderError(e) from None
        except PendingActionUnresolvedOutcomeError:
            # Do not trace PendingActionUnresolvedOutcomeError
            raise
        except ActionRenderError as e:
            self.logger.debug(f"Rendering {value!r} failed: {e!r}", exc_info=True)
            raise

    def _internal_render_string(self, value: str) -> str:
        """Recursive rendering routine"""
        self._depth += 1
        if self._depth >= MAX_RECURSION_DEPTH:
            # This exception floats to the very "render" call without any logging
            raise ActionRenderRecursionError(f"Recursion depth exceeded: {self._depth}/{MAX_RECURSION_DEPTH}")
        try:
            chunks: list[str] = []
            # Cheap check
            if not qualify_string_as_potentially_renderable(value):
                return value
            for lexeme_type, lexeme_value in TemplarStringLexer(value):
                if lexeme_type == TemplarStringLexer.EXPRESSION:
                    lexeme_value = str(self._eval(expression=lexeme_value))
                chunks.append(lexeme_value)
            return "".join(chunks)
        finally:
            self._depth -= 1

    @classmethod
    def _make_restricted_builtin_call_shim(cls, name: str) -> t.Callable:
        def _call(*args, **kwargs) -> t.NoReturn:
            raise RestrictedBuiltinError(name)

        return _call

    def _eval(self, expression: str) -> t.Any:
        """Safely evaluate an expression."""
        self.logger.debug(f"Processing expression: {expression!r}")
        try:
            # pylint: disable=eval-used
            return eval(expression, self._globals, {})  # nosec
        except ActionRenderError:
            raise
        except Exception as e:
            render_failure_source: str = str(e) if isinstance(e, (SyntaxError, NameError)) else repr(e)
            self.logger.warning(f"Expression render failed: {e!r} (for {expression!r})")
            raise ActionRenderError(render_failure_source) from e

    def render(self, data: t.Any) -> t.Any:
        """Perform recursive rendering"""
        result: t.Any
        if isinstance(data, dict):
            result = {k: self.render(v) for k, v in data.items()}
        elif isinstance(data, list):
            result = [self.render(v) for v in data]
        elif isinstance(data, str):
            result = self._render_string(data)
        elif isinstance(data, Expression):
            evaluated_expression: t.Any = self._eval(data.expression)
            result = self.render(evaluated_expression)
        else:
            result = data
        # Unwrap lazy proxies
        while isinstance(result, c.LazyProxy):
            result = result.__wrapped__
        return result


class WorkflowTemplar(CommonTemplar):
    """Expression renderer specifically for workflows"""

    def __init__(
        self,
        *,
        outcomes_map: t.Mapping[str, t.Mapping[str, str]],
        action_states: t.Mapping[str, str],
        context_map: t.Mapping[str, t.Any],
        locals_map: t.Mapping[str, t.Any],
        metadata: t.Optional[t.Mapping[str, t.Any]] = None,
    ) -> None:
        outcomes_container: c.AttrDict = c.ActionOutcomeAggregateDict()
        for name, status_str in action_states.items():
            action_outcomes: t.Optional[dict] = None
            if status_str not in (ActionStatus.PENDING.value, ActionStatus.RUNNING.value):
                action_outcomes = c.OutcomeDict(outcomes_map.get(name, {}))
            outcomes_container[name] = action_outcomes
        status_container: c.AttrDict = c.ActionContainingDict(action_states)
        context_container: c.AttrDict = c.AttrDict({k: self._load_ctx_node(data=v) for k, v in context_map.items()})
        locals_container: c.AttrDict = c.AttrDict({k: self._load_ctx_node(data=v) for k, v in locals_map.items()})
        environment_container: c.AttrDict = c.LooseDict(os.environ)
        metadata_container: c.AttrDict = c.LooseDict({"status": status_container})
        modules_container: c.AttrDict = c.ExternalModulesDict()
        if metadata is not None:
            metadata_container.update(metadata)
        super().__init__(
            # Full names
            outcomes=outcomes_container,
            context=context_container,
            environment=environment_container,
            locals=locals_container,
            metadata=metadata_container,
            modules=modules_container,
            # Aliases
            out=outcomes_container,
            ctx=context_container,
            env=environment_container,
            loc=locals_container,
            meta=metadata_container,
            mod=modules_container,
        )

    def _evaluate_context_object_expression(self, expression: str) -> t.Any:
        obj: t.Any = self._eval(expression)
        return self._load_ctx_node(obj)

    def _load_ctx_node(self, data: t.Any) -> t.Any:
        """Deep copy of context data,
        while transforming dicts into attribute-accessor proxies
        and turning leaf string values into deferred templates."""
        if isinstance(data, Expression):
            return c.LazyProxy(lambda: self._evaluate_context_object_expression(data.expression))
        if isinstance(data, dict):
            result_dict = c.AttrDict()
            for key, value in data.items():
                result_dict[key] = self._load_ctx_node(value)
            return result_dict
        if isinstance(data, list):
            result_list = []
            for item in data:
                result_list.append(self._load_ctx_node(item))
            return result_list
        if isinstance(data, str) and qualify_string_as_potentially_renderable(data):
            return c.LazyProxy(lambda: self._internal_render_string(data))
        return data



# pylint: disable=invalid-field-call
"""Separate module for shell-related action"""

import dataclasses
import functools
import os
import pathlib
import shlex
import typing as t
from asyncio.subprocess import create_subprocess_shell, Process  # noqa
from subprocess import PIPE  # nosec

from ..base import ArgsBase, CaptureStream, StreamCaptureConfiguration, SubprocessActionBase
from ...config.constants import C

__all__ = [
    "ShellAction",
]


class ShellArgsByCommand(ArgsBase):
    """Args for shell-related actions with a command provided"""

    command: str
    environment: t.Optional[dict[str, str]] = None
    cwd: t.Optional[str] = None
    executable: t.Optional[str] = None
    capture: list[CaptureStream] = dataclasses.field(default_factory=list)


class ShellArgsByFile(ArgsBase):
    """Args for shell-related actions with a file provided"""

    file: pathlib.Path
    environment: t.Optional[dict[str, str]] = None
    cwd: t.Optional[str] = None
    executable: t.Optional[str] = None
    capture: list[CaptureStream] = dataclasses.field(default_factory=list)


class ShellAction(SubprocessActionBase):
    """Runs a shell command on the local system."""

    args: t.Union[ShellArgsByCommand, ShellArgsByFile]

    @functools.cache
    def _get_capture_configration(self) -> StreamCaptureConfiguration:
        return StreamCaptureConfiguration.from_streams_list(self.args.capture)

    async def _create_process(self) -> Process:
        command: str
        if isinstance(self.args, ShellArgsByCommand):
            command = self.args.command
        else:
            command = f". {shlex.quote(str(self.args.file))}"
        if C.SHELL_INJECT_YIELD_FUNCTION:
            command = f"{self._SHELL_SERVICE_FUNCTIONS_DEFINITIONS}\n{command}"
        environment: t.Optional[dict[str, str]] = None
        if self.args.environment is not None:
            environment = os.environ.copy()
            environment.update(self.args.environment)
        process = await create_subprocess_shell(
            cmd=command,
            stdin=PIPE,
            stdout=PIPE,
            stderr=PIPE,
            env=environment,
            cwd=self.args.cwd,
            executable=self.args.executable or C.DEFAULT_SHELL_EXECUTABLE,
            limit=C.SUBPROCESS_STREAM_BUFFER_LIMIT,
        )
        return process



"""Dataclass loading exceptions"""

import dataclasses
import typing as t

from ..concealment import represent_object_type


class ClassLoaderError(Exception):
    """Base load error"""

    def __init__(self, path: str, message: str) -> None:
        self.path: str = path or "/"
        self.message: str = message
        info: str = self.path if self.message is None else f"{self.message} (at {self.path!r})"
        super().__init__(info)


class TypeMatchError(ClassLoaderError):
    """Type mismatches"""

    def __init__(self, data_type: type, value: t.Any, path: str) -> None:
        value_type_str: str = represent_object_type(value)
        message: str = f"Unrecognized content type: {value_type_str} (expected {data_type!r})"
        super().__init__(path=path, message=message)


class MissingValueError(ClassLoaderError):
    """Missing value for field"""

    def __init__(self, path: str) -> None:
        message: str = "Missing value for field"
        super().__init__(path=path, message=message)


class UnionTypeMatchError(TypeMatchError):
    """None of the union types match"""


class StrictUnionTypeMatchError(ClassLoaderError):
    """More than one of the union types match"""

    def __init__(self, matches: dict[type, t.Any], path: str) -> None:
        match_types = ", ".join(sorted(str(data_type) for data_type in matches))
        message: str = f"Multiple union matches found: {match_types}"
        super().__init__(path=path, message=message)


class UnexpectedDataError(ClassLoaderError):
    """Unexpected fields"""

    def __init__(self, keys: set[str], path: str) -> None:
        message: str = f"Unrecognized fields: {', '.join(repr(k) for k in sorted(keys))}"
        super().__init__(path=path, message=message)


class RootTypeUnionMatchError(ClassLoaderError):
    """Indicates that the union-based argument spec did not match"""

    def __init__(self, types: tuple, keys: list[str]) -> None:
        type_signatures: list[str] = []
        for typ in types:
            type_field_names: list[str] = sorted(f.name for f in dataclasses.fields(typ))
            type_signatures.append(f"{type_field_names} for {typ.__name__}")
        expected: str = ", ".join(type_signatures)
        message: str = f"Data did not conform to allowed type signatures: {keys}, expected one of: {expected}"
        super().__init__(path="", message=message)
